package cn.jlsxwkj.moudles.chat;

import cn.dev33.satoken.SaManager;
import cn.hutool.core.lang.Tuple;
import cn.hutool.crypto.digest.MD5;
import cn.hutool.json.JSONObject;
import cn.hutool.json.JSONUtil;
import cn.jlsxwkj.common.config.UserConfig;
import cn.jlsxwkj.common.exception.CustomException;
import cn.jlsxwkj.common.exception.FileTypeDoesNotSupportException;
import cn.jlsxwkj.common.exception.InsertFailException;
import cn.jlsxwkj.common.exception.OpenFileFailException;
import cn.jlsxwkj.common.exception.UnknownException;
import cn.jlsxwkj.common.utils.FileType;
import cn.jlsxwkj.common.utils.Log;
import cn.jlsxwkj.common.utils.MergeDocuments;
import cn.jlsxwkj.moudles.chat_history.ChatHistoryService;
import jakarta.annotation.Resource;
import org.springframework.ai.chat.messages.Message;
import org.springframework.ai.chat.messages.SystemMessage;
import org.springframework.ai.chat.messages.UserMessage;
import org.springframework.ai.chat.prompt.Prompt;
import org.springframework.ai.document.Document;
import org.springframework.ai.document.DocumentReader;
import org.springframework.ai.embedding.EmbeddingModel;
import org.springframework.ai.ollama.OllamaChatModel;
import org.springframework.ai.ollama.api.OllamaOptions;
import org.springframework.ai.vectorstore.SearchRequest;
import org.springframework.ai.vectorstore.VectorStore;
import org.springframework.stereotype.Service;
import org.springframework.web.multipart.MultipartFile;
import reactor.core.publisher.Flux;

import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Locale;
import java.util.Objects;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.stream.Collectors;

import static cn.jlsxwkj.common.config.FunctionCallConfig.TOOLS_FUNCTION;

/**
 * @author zh
 */
@Service
public class ChatService {

    /**
     * md5.
     */
    private final MD5 md5 = cn.hutool.crypto.digest.MD5.create();
    /**
     * 获取当前项目地址 + /ai_doc.
     */
    private static final String PATH = System.getProperty("user.dir") + "/ai_doc/";
    /**
     * pg 向量对象(将文档转换为向量并保存至数据库).
     */
    @Resource
    private VectorStore vectorStore;
    /**
     * ollama 模型.
     */
    @Resource
    private OllamaChatModel ollamaChatModel;
    /**
     * 向量化模型.
     */
    @Resource
    private EmbeddingModel embeddingModel;
    /**
     * 保存对话对象(将对话保存至数据库).
     */
    @Resource
    private ChatHistoryService chatHistoryService;
    /**
     * 自定义配置类.
     */
    @Resource
    private UserConfig userConfig;

    /**
     * 使用spring ai解析txt文档并保存至 pg.
     *
     * @param file 保存的文件
     * @return 保存状态
     */
    public String uploadDocument(
        final MultipartFile file) throws CustomException {
        Tuple supportFile = isSupportFile(file);
        File saveFile = supportFile.get(0);
        String fileName = supportFile.get(1);
        DocumentReader reader = supportFile.get(3);
        if (!saveFile.exists()) {
            try {
                file.transferTo(saveFile);
                vectorStore.add(reader.get());
            } catch (Exception e) {
                boolean delete = saveFile.delete();
                if (!delete) {
                    Log.warn(this.getClass(), "删除文件失败: {}", fileName);
                    throw new UnknownException();
                }
            }
        }
        return "上传完成";
    }

    /**
     * 根据关键词搜索向量库.
     *
     * @param keyword 关键词
     * @return 文本内容
     */
    public String search(final String keyword) {
        SearchRequest searchRequest = SearchRequest.query(keyword).withSimilarityThreshold(0.5);
        List<Document> documents = vectorStore.similaritySearch(searchRequest);
        return mergeDocuments(documents);
    }

    /**
     * 问答流,根据输入内容回答.
     *
     * @param message 输入内容
     * @return 回答内容
     */
    public Flux<String> chatStream(final String message) {
        String context = search(message);
        Prompt prompt = new Prompt(List.of(
            new SystemMessage(userConfig.getSysMessage()),
            new UserMessage(message)),
            OllamaOptions.builder().withFunctions(TOOLS_FUNCTION).build());
        AtomicBoolean flag = new AtomicBoolean(false);
        StringBuilder funcCall = new StringBuilder();
        String funcCallResult = "";
        Objects.requireNonNull(ollamaChatModel.stream(prompt)
            .map(response -> {
                String content = response.getResult().getOutput().getContent();
                if (content.contains("<tool")) {
                    flag.set(true);
                }
                return content;
            }).collectList().block())
            .forEach(funcCall::append);
        if (flag.get()) {
            String tool = funcCall.toString().split("\n")[1];
            JSONObject entries = JSONUtil.parseObj(tool);
            String funcName = entries.get("name", String.class);
            String args = entries.get("arguments", String.class);
            funcCallResult = ollamaChatModel.getFunctionCallbackRegister().get(funcName).call(args);
        }
        return stream(new UserMessage(message),
            new SystemMessage("""
                %s
                %s
                根据参考内容回答, 当内容和问题不匹配时不要参考 system context, 请参考 user context 回答
                """.formatted(context, funcCallResult))
        );
    }

    /**
     * 问答总结流, 根据给定的文档总结内容.
     *
     * @param message 输入内容
     * @param file    文件
     * @return 内容总结
     */
    public Flux<String> chatSummaryStream(final String message,
                                          final MultipartFile file) {
        embeddingModel.embed("");
        Tuple supportFile;
        try {
            supportFile = isSupportFile(file);
        } catch (CustomException e) {
            return Flux.just(e.getMessage()).concatWithValues("<{完成}>");
        }
        String fileName = supportFile.get(1);
        String fileTypeCn = supportFile.get(2);
        DocumentReader reader = supportFile.get(3);
        String document = mergeDocuments(reader.get());
        return stream(new UserMessage(message),
            new UserMessage("""
                %s :: %s ====> %s
                ↑ %s
                """.formatted(fileTypeCn, fileName, document, message))
        );
    }

    /**
     * 合并文档.
     *
     * @param documents 文档列表
     * @return 合并后的文档
     */
    private String mergeDocuments(final List<Document> documents) {
        return MergeDocuments.mergeDocuments(documents).stream()
            .map(Document::getContent).collect(Collectors.joining("\n"));
    }


    /**
     * chat stream 封装.
     *
     * @param userMessage 提问内容
     * @param sysOrUserMessage 用户或系统提示词
     * @return flux
     */
    private Flux<String> stream(final Message userMessage,
                                final Message sysOrUserMessage) {
        String userId = SaManager.getStpLogic("").getLoginIdAsString();
        StringBuilder chatMessage = new StringBuilder();
        return ollamaChatModel.stream(sysOrUserMessage, userMessage).map(response -> {
            chatMessage.append(response);
            return response.replace(" ", "  ").replace("\t", "    ");
        }).concatWithValues("<{完成}>")
            .doOnComplete(() -> {
                String chatMsg = chatMessage.toString();
                try {
                    chatHistoryService.insert(userId, userMessage.getContent(), chatMsg);
                } catch (InsertFailException e) {
                    Log.warn(this.getClass(), e.getMessage());
                }
            });
    }

    /**
     * 封装文件.
     *
     * @param file 文件流
     * @return 元组
     */
    private Tuple isSupportFile(
        final MultipartFile file) throws CustomException {
        byte[] bytes;
        try {
            bytes = file.getBytes();
        } catch (IOException e) {
            throw new OpenFileFailException("读取文件失败");
        }
        String[] split = Objects.requireNonNull(file.getOriginalFilename()).split("\\.");
        String fileType = split[split.length - 1].toLowerCase(Locale.ROOT);
        File savePath = new File(PATH);
        File saveFile = new File(PATH + md5.digestHex(bytes) + "." + fileType);
        String fileName = saveFile.getName();
        if (!savePath.exists()) {
            if (!savePath.mkdirs()) {
                Log.warn(this.getClass(), "创建文件夹失败: " + PATH);
            }
        }
        Tuple fileTypeAndResource = new FileType(bytes, fileName, 5, userConfig.getCnocrUrl())
            .getFileTypeAndResource(fileType);
        if (fileTypeAndResource == null) {
            throw new FileTypeDoesNotSupportException("暂不支持的文件类型: " + fileType);
        }
        String fileTypeCn = fileTypeAndResource.get(0);
        DocumentReader reader = fileTypeAndResource.get(1);
        return new Tuple(saveFile, fileName, fileTypeCn, reader);
    }
}
